(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2020 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

Quiet[
    BeginPackage["AtomicDensityMatrix`AtomicData`", {
        "ProcessOptions`",
        "PhysicalConstants`",
        "Units`",
        "AtomicDensityMatrix`",
        "AtomicDensityMatrix`Common`"}],
    General::obspkg
];

Begin["`Private`"];

$AtomicDataFields = {
    ElementName, ElementSymbol, AtomicNumber, AtomicWeight, Abundance, NuclearSpin,
    J, L, S, NaturalWidth, Energy, HyperfineA, HyperfineB, Parity, Wavenumber,
    Lifetime, BranchingRatio, PartialWidth
};

AtomicDataFieldsPattern[] = Alternatives @@ $AtomicDataFields;

AtomicData::alternates =
    "Requested state not in database.  Possible alternatives: `1`";

Options[AtomicData] = {StateLabel -> None};

OptionProcessor[StateLabel] = Replace[#, None -> False]&;

AtomicData[field : AtomicDataFieldsPattern[]] :=
    Select[Map[#[[1, 1, 1]]&, DownValues@field], ListQ[#] && Length[#] > 2 &];

AtomicData[] := Union[Flatten[AtomicData /@ $AtomicDataFields, 1]];

AtomicData[state_List] /; Length[state] < 4 :=
    Select[AtomicData[], Take[#, Length[state]] === state&];

AtomicData[state_List, opts : OptionsPattern[]] :=
    Module[{data},
        data = Map[AtomicData[state, #, opts]&, $AtomicDataFields];
        data = Select[data, Not@MemberQ[$AtomicDataFields, Head@Last@#]&]
    ];

(AtomicData[state_List, field : AtomicDataFieldsPattern[], opts : OptionsPattern[]] := Module[
    {stateLabel},
    If[FalseQ@stateLabel,
        field -> field@state,
        field@stateLabel -> field@state
    ]
]) // ProcessOptions;

(AtomicData[state2_List, (field : BranchingRatio | PartialWidth)[state1_List], opts : OptionsPattern[]] := Module[
    {stateLabel},
    notation = Which[
        FalseQ@stateLabel,
        field[state1],
        TrueQ@stateLabel,
        field[state2, state1],
        FunctionDefinedQ[stateLabel, state1],
        field[stateLabel[state2], stateLabel[state1]],
        True,
        field[stateLabel, state1]
    ];
    notation -> field[state1][state2]
]) // ProcessOptions;
AtomicData[state_List, list_List, opts : OptionsPattern[]] := AtomicData[state, #, opts]& /@ list;

AtomicDataDefinedQ[state_List, field : AtomicDataFieldsPattern[]] := field@state =!= Unevaluated[field@state];

Options[AtomicSystem] = {StateLabels -> Automatic, DropInexactValues -> False};

OptionProcessor[StateLabels] = Switch[#,
    Automatic,
    Function[states, Range@Length@states],
    OptionsPattern[],
    Function[states, states /. #],
    _List,
    Function[states, PadRight[{}, Length@states, #]],
    _,
    Function[states, # /@ states]
]&;

InexactExpressionPattern[] = expr_ /; !FreeQ[expr, n_ /; InexactNumberQ[n]];

(AtomicSystem[states_, parameters_, opts : OptionsPattern[]] := Module[
    {stateLabels, dropInexactValues, fullParameters, system},
    fullParameters = Flatten[parameters /. BranchingRatio -> BranchingRatio /@ states];
    system = AtomicState[#, AtomicData[#, fullParameters]]& /@ states /. Thread[states -> stateLabels@states];
    If[dropInexactValues, system = DeleteCases[system, _ -> InexactExpressionPattern[], Infinity]];
    system
]) // ProcessOptions;

(* Atomic Information *)

AtomicElements = {
    {"H", 1, "Hydrogen"}, {"He", 2, "Helium"}, {"Li", 3, "Lithium"}, {"Be", 4, "Beryllium"}, {"B", 5, "Boron"},
    {"C", 6, "Carbon"}, {"N", 7, "Nitrogen"}, {"O", 8, "Oxygen"}, {"F", 9, "Fluorine"}, {"Ne", 10, "Neon"},
    {"Na", 11, "Sodium"}, {"Mg", 12, "Magnesium"}, {"Al", 13, "Aluminium"}, {"Si", 14, "Silicon"},
    {"P", 15, "Phosphorus"}, {"S", 16, "Sulfur"}, {"Cl", 17, "Chlorine"}, {"Ar", 18, "Argon"}, {"K", 19, "Potassium"},
    {"Ca", 20, "Calcium"}, {"Sc", 21, "Scandium"}, {"Ti", 22, "Titanium"}, {"V", 23, "Vanadium"},
    {"Cr", 24, "Chromium"}, {"Mn", 25, "Manganese"}, {"Fe", 26, "Iron"}, {"Co", 27, "Cobalt"}, {"Ni", 28, "Nickel"},
    {"Cu", 29, "Copper"}, {"Zn", 30, "Zinc"}, {"Ga", 31, "Gallium"}, {"Ge", 32, "Germanium"}, {"As", 33, "Arsenic"},
    {"Se", 34, "Selenium"}, {"Br", 35, "Bromine"}, {"Kr", 36, "Krypton"}, {"Rb", 37, "Rubidium"},
    {"Sr", 38, "Strontium"}, {"Y", 39, "Yttrium"}, {"Zr", 40, "Zirconium"}, {"Nb", 41, "Niobium"},
    {"Mo", 42, "Molybdenum"}, {"Tc", 43, "Technetium"}, {"Ru", 44, "Ruthenium"}, {"Rh", 45, "Rhodium"},
    {"Pd", 46, "Palladium"}, {"Ag", 47, "Silver"}, {"Cd", 48, "Cadmium"}, {"In", 49, "Indium"}, {"Sn", 50, "Tin"},
    {"Sb", 51, "Antimony"}, {"Te", 52, "Tellurium"}, {"I", 53, "Iodine"}, {"Xe", 54, "Xenon"}, {"Cs", 55, "Caesium"},
    {"Ba", 56, "Barium"}, {"La", 57, "Lanthanum"}, {"Ce", 58, "Cerium"}, {"Pr", 59, "Praseodymium"},
    {"Nd", 60, "Neodymium"}, {"Pm", 61, "Promethium"}, {"Sm", 62, "Samarium"}, {"Eu", 63, "Europium"},
    {"Gd", 64, "Gadolinium"}, {"Tb", 65, "Terbium"}, {"Dy", 66, "Dysprosium"}, {"Ho", 67, "Holmium"},
    {"Er", 68, "Erbium"}, {"Tm", 69, "Thulium"}, {"Yb", 70, "Ytterbium"}, {"Lu", 71, "Lutetium"},
    {"Hf", 72, "Hafnium"}, {"Ta", 73, "Tantalum"}, {"W", 74, "Tungsten"}, {"Re", 75, "Rhenium"}, {"Os", 76, "Osmium"},
    {"Ir", 77, "Iridium"}, {"Pt", 78, "Platinum"}, {"Au", 79, "Gold"}, {"Hg", 80, "Mercury"}, {"Tl", 81, "Thallium"},
    {"Pb", 82, "Lead"}, {"Bi", 83, "Bismuth"}, {"Po", 84, "Polonium"}, {"At", 85, "Astatine"}, {"Rn", 86, "Radon"},
    {"Fr", 87, "Francium"}, {"Ra", 88, "Radium"}, {"Ac", 89, "Actinium"}, {"Th", 90, "Thorium"},
    {"Pa", 91, "Protactinium"}, {"U", 92, "Uranium"}, {"Np", 93, "Neptunium"}, {"Pu", 94, "Plutonium"},
    {"Am", 95, "Americium"}, {"Cm", 96, "Curium"}, {"Bk", 97, "Berkelium"}, {"Cf", 98, "Californium"},
    {"Es", 99, "Einsteinium"}, {"Fm", 100, "Fermium"}, {"Md", 101, "Mendelevium"}, {"No", 102, "Nobelium"},
    {"Lr", 103, "Lawrencium"}, {"Rf", 104, "Rutherfordium"}, {"Db", 105, "Dubnium"}, {"Sg", 106, "Seaborgium"},
    {"Bh", 107, "Bohrium"}, {"Hs", 108, "Hassium"}, {"Mt", 109, "Meitnerium"}
};

Do[
    ElementName[element[[1]]] = element[[3]];
    AtomicNumber[element[[1]]] = element[[2]],
    {element, AtomicElements}
];

LSymbols = {"S", "P", "D", "F", "G", "H", "I", "K", "L", "M", "N", "O", "Q", "R", "T", "U"};

StateDesignationPattern[] = {
    name_String,
    weight_Integer | {weight_Integer, isomer_String},
    configuration : {
        core_String | electron : {_Integer, _String} | {_Integer, _String, _Integer},
        electrons : ({_Integer, _String} | {_Integer, _String, _Integer})...
    },
    term : {multiplicity_Integer, l_String, j : (_Rational | _Integer)}
};

ElementSymbol[state : StateDesignationPattern[]] := state /. StateDesignationPattern[] -> name;
AtomicWeight[state : StateDesignationPattern[]] := state /. StateDesignationPattern[] -> weight;
CoreSymbol[state : StateDesignationPattern[]] := state /. StateDesignationPattern[] -> core;
ValenceElectrons[state : StateDesignationPattern[]] := state /. StateDesignationPattern[] -> {electron, electrons};
SpinMultiplicity[state : StateDesignationPattern[]] := state /. StateDesignationPattern[] -> multiplicity;
LSymbol[state : StateDesignationPattern[]] := state /. StateDesignationPattern[] -> l;
J[state : StateDesignationPattern[]] := state /. StateDesignationPattern[] -> j;
AtomicConfiguration[state : StateDesignationPattern[]] := state /. StateDesignationPattern[] -> configuration;
AtomicTerm[state : StateDesignationPattern[]] := state /. StateDesignationPattern[] -> term;
TermSymbol[state : StateDesignationPattern[]] := Style[
    Row@{Superscript["", #1], Subscript[#2, InputForm@#3]} & @@ AtomicTerm@@state,
    "InlineFormula", ScriptMinSize -> 3
];

GetStateData[state : StateDesignationPattern[]] := (
    ElementSymbol[state] = ElementSymbol@state;
    ElementName[state] = ElementName@ElementSymbol@state;
    AtomicNumber[state] = AtomicNumber@ElementSymbol@state;
    AtomicWeight[state] = AtomicWeight@state;
    J[state] = J@state;
    LSymbol[state] = LSymbol@state;
    L[state] = Position[LSymbols, LSymbol@state, {1}, Heads -> False][[1, 1]] - 1;
    SpinMultiplicity[state] = SpinMultiplicity@state;
    S[state] = (SpinMultiplicity@state - 1)/2;
    Abundance[state] = Abundance[Take[state, 2]];
    NuclearSpin[state] = NuclearSpin[Take[state, 2]];
    AtomicTerm[state] = AtomicTerm@state;
    TermSymbol[state] = TermSymbol@state;
);

ElectronDesignationPattern[] = {principalNumber_Integer, l_String} | {principalNumber_Integer, l_String, equivalency_Integer};

PrincipalQuantumNumber[electron : ElectronDesignationPattern[]] := electron /. ElectronDesignationPattern[] -> principalNumber;
ElectronL[electron : ElectronDesignationPattern[]] := electron /. ElectronDesignationPattern[] -> l;
ElectronEquivalency[electron : ElectronDesignationPattern[]] := electron /. ElectronDesignationPattern[] -> equivalency;

EnergyFromWavenumber[state_] := 2\[Pi] Convert[SpeedOfLight Wavenumber[state], Mega/Second];

NaturalWidthFromLifetime[state_] := Convert[1/Lifetime[state], Mega/Second];

LifetimeFromNaturalWidth[state_] := Convert[1/NaturalWidth[state], Nano Second];

PossibleDecays[state_] := Module[{wavenumber, parity, j},
    {wavenumber, parity, j} = Through@{Wavenumber, Parity, J}[state] /. Centimeter -> 1;
    Select[
        AtomicData[Take[state, 2]],
        Centimeter Wavenumber[#] < wavenumber && Parity[#] =!= parity && Abs[J[#] - j] <= 1 &
    ]
];

NaturalWidthFromPartialWidths[state_] := Total@Through[(PartialWidth /@ PossibleDecays[state])[state]];

DefineBranchingRatiosFromPartialWidths[state_] :=
    (BranchingRatio[#][state] = Convert[PartialWidth[#][state] / NaturalWidth[state], 1])& /@ PossibleDecays[state];

DopplerWidth[groundState_, excitedState_, t_] :=
    2 Pi Convert[
        (Wavenumber@excitedState - Wavenumber@groundState) MostProbableSpeed[groundState, t],
        Mega 1/Second
    ];

MostProbableSpeed[state_, t_] :=
    Convert[
        Sqrt[(2 BoltzmannConstant t Kelvin)/(AtomicWeight@state AtomicMassUnit)],
        Centimeter/Second
    ];

VaporPressureFunction[range_List, a_List, b_List, c_:{0}, d_:{0}][t_] := Piecewise[
    MapThread[
        {VaporPressureFunction[#1, #2, #3, #4][t]/Atmosphere, #5 < t < #6}&,
        {a, b, PadRight[c, Length@a], PadRight[d, Length@a], Most@range, Rest@range}
    ]
] Atmosphere;

VaporPressureFunction[a:Except[_List], b:Except[_List], c_:0, d_:0][t_] :=
    10^(a + b/t + c Log[10, t] + d/t^3) Atmosphere;

VaporPressure[state_ /; Not@VectorQ@state, t_] :=
    VaporPressure[TakeWhile[state, !ListQ@#&], t];

AtomicDensity[state_, t_] := Convert[
    VaporPressure[state, t]/(BoltzmannConstant t Kelvin),
    1/Centimeter^3
];

BranchingRatio[state1_][state2_] /; (Parity[state1] == Parity[state2] || Energy[state1] > Energy[state2] || Abs[J[state1] - J[state2]] > 1 /. {Mega -> 10^6, Second -> 1}) = 0;

PartialWidth[state1_][state2_] := BranchingRatio[state1][state2] NaturalWidth[state2];

(*
	Na, Rb, and Cs data from http://steck.us/alkalidata/
	Rb natural abundance from http://www.tracesciences.com/rb.htm
 *)

(* H 

Wavenumbers from http://www.nist.gov/physlab/data/handbook/index.cfm

*)

state = {"H", 1};
Abundance[state] = 1;
NuclearSpin[state] = 1/2;

(state = {"H", 1, {{1, "s"}}, {2, "S", 1/2}}) // GetStateData;
Lifetime[state] = Infinity Second;
Wavenumber[state] = 0;
HyperfineA[state] = 2\[Pi] 1420. Mega 1/Second;
HyperfineB[state] = 0;
Parity[state] = Even;

(state = {"H", 1, {{2, "p"}}, {2, "P", 1/2}}) // GetStateData;
Lifetime[state] = 1.60 Nano Second;
Wavenumber[state] = 82258.9191 1/Centimeter;
HyperfineA[state] = 0 2\[Pi] Mega 1/Second;
HyperfineB[state] = 0;
Parity[state] = Odd;

(state = {"H", 1, {{2, "p"}}, {2, "P", 3/2}}) // GetStateData;
Lifetime[state] = 1.60 Nano Second;
Wavenumber[state] = 82259.2850 1/Centimeter;
HyperfineA[state] = 0 2\[Pi] Mega 1/Second;
HyperfineB[state] = 0;
Parity[state] = Odd;

(* He *)

state = {"He", 2};
Abundance[state] = 1;
NuclearSpin[state] = 0;

(state = {"He", 2, {{1, "s", 2}}, {1, "S", 0}}) // GetStateData;
Lifetime[state] = Infinity Second;
Wavenumber[state] = 0;
HyperfineA[state] = 0;
HyperfineB[state] = 0;
Parity[state] = Even;

(state = {"He", 2, {{1, "s"}, {2, "s"}}, {3, "S", 1}}) // GetStateData;
Wavenumber[state] = 159855.9743297 1/Centimeter;
HyperfineA[state] = 0;
HyperfineB[state] = 0;
Parity[state] = Even;

(state = {"He", 2, {{1, "s"}, {2, "s"}}, {1, "S", 0}}) // GetStateData;
Wavenumber[state] = 166277.440141 1/Centimeter;
HyperfineA[state] = 0;
HyperfineB[state] = 0;
Parity[state] = Even;

(state = {"He", 2, {{1, "s"}, {2, "p"}}, {3, "P", 2}}) // GetStateData;
Lifetime[state] = Convert[1 / (2 Pi 1.62 Mega/Second), Nano Second];  (* [Mv99] *)
Wavenumber[state] = 169086.7664725 1/Centimeter;
HyperfineA[state] = 0;
HyperfineB[state] = 0;
Parity[state] = Odd;

(state = {"He", 2, {{1, "s"}, {2, "p"}}, {3, "P", 1}}) // GetStateData;
Lifetime[state] = Convert[1 / (2 Pi 1.62 Mega/Second), Nano Second]; (* [Mv99] *)
Wavenumber[state] = 169086.8428979 1/Centimeter;
HyperfineA[state] = 0;
HyperfineB[state] = 0;
Parity[state] = Odd;

(state = {"He", 2, {{1, "s"}, {2, "p"}}, {3, "P", 0}}) // GetStateData;
Lifetime[state] = Convert[1 / (2 Pi 1.62 Mega/Second), Nano Second]; (* [Mv99] *)
Wavenumber[state] = 169087.8308131 1/Centimeter;
HyperfineA[state] = 0;
HyperfineB[state] = 0;
Parity[state] = Odd;

(* Li *)

state = {"Li", 3};
Abundance[state] = 92.41;
NuclearSpin[state] = 3/2;

(state = {"Li", 3, {{1, "s", 2}, {2, "s"}}, {2, "S", 1/2}}) // GetStateData;
Lifetime[state] = Infinity Second;
Wavenumber[state] = 0;
HyperfineA[state] = 401.7520433 (2 Pi) Mega 1/Second; (* [AIV77] *)
HyperfineB[state] = 0;
Parity[state] = Even;

(state = {"Li", 3, {{1, "s", 2}, {2, "p"}}, {2, "P", 1/2}}) // GetStateData;
Lifetime[state] = 27.2 Nano Second;
Wavenumber[state] = 14903.66 1/Centimeter;
HyperfineA[state] = 46.024 (2 Pi) Mega 1/Second; (* [DN08] *)
HyperfineB[state] = 0;
Parity[state] = Odd;

(state = {"Li", 3, {{1, "s", 2}, {2, "p"}}, {2, "P", 3/2}}) // GetStateData;
Lifetime[state] = 27.2 Nano Second;
Wavenumber[state] = 14904.00 1/Centimeter;
HyperfineA[state] = -3.055 (2 Pi) Mega 1/Second; (* [AIV77] *)
HyperfineB[state] = -0.221 (2 Pi) Mega 1/Second; (* [AIV77] *)
Parity[state] = Odd;

(* Na *)

state = {"Na", 23};
Abundance[state] = 1;
NuclearSpin[state] = 3/2;
AtomicWeight[state] = 22.9897692807 (* [BPR+99] *);
MeltingPoint[state] = 97.794 + 273.15; (* [Lin05] *)
BoilingPoint[state] = 882.940 + 273.15; (* [Lin05] *)
VaporPressure[state, t_] = VaporPressureFunction[{298, MeltingPoint@state, 700}, {5.298, 4.704}, {-5603, -5377}][t]; (* [AIH84] *)

(state = {"Na", 23, {"Ne", {3, "s"}}, {2, "S", 1/2}}) // GetStateData;
Parity[state] = Even;
Wavenumber[state] = 0;
HyperfineA[state] = 2\[Pi] 885.81306440 Mega 1/Second; (* [San08] *)
HyperfineB[state] = 0;
Lifetime[state] = Infinity Second;

(state = {"Na", 23, {"Ne", {3, "p"}}, {2, "P", 1/2}}) // GetStateData;
Parity[state] = Odd;
Wavenumber[state] = 16956.170250 1/Centimeter; (* [San08] *)
HyperfineA[state] = 2\[Pi] 94.44 Mega 1/Second; (* [San08] *)
HyperfineB[state] = 0;
Lifetime[state] = 16.299 Nano Second;
BranchingRatio[{"Na", 23, {"Ne", {3, "s"}}, {2, "S", 1/2}}][state] = 1;

(state = {"Na", 23, {"Ne", {3, "p"}}, {2, "P", 3/2}}) // GetStateData;
Parity[state] = Odd;
Wavenumber[state] = 16973.366160 1/Centimeter; (* [San08] *)
HyperfineA[state] = 2\[Pi] 18.534 Mega 1/Second; (* [San08] *)
HyperfineB[state] = 2\[Pi] 2.724 Mega 1/Second; (* [San08] *)
Lifetime[state] = 16.2492 Nano Second;
BranchingRatio[{"Na", 23, {"Ne", {3, "s"}}, {2, "S", 1/2}}][state] = 1;

(state = {"Na", 23, {"Ne", {3, "d"}}, {2, "D", 3/2}}) // GetStateData;
Parity[state] = Even;
Wavenumber[state] = 29172.889 1/Centimeter; (* [KP08] *)
HyperfineA[state] = 2\[Pi] 0.527 Mega 1/Second; (* [BHM88] *)
HyperfineB[state] = 2\[Pi] 0.126 Mega 1/Second; (* [SDJ98] *)
PartialWidth[{"Na", 23, {"Ne", {3, "p"}}, {2, "P", 1/2}}][state] = 42.9 Mega/Second; (* [KP08, Hea61] *)
PartialWidth[{"Na", 23, {"Ne", {3, "p"}}, {2, "P", 3/2}}][state] = 8.57 Mega/Second; (* [KP08, Hea61] *)
NaturalWidth[state] = NaturalWidthFromPartialWidths[state];
DefineBranchingRatiosFromPartialWidths[state];

(state = {"Na", 23, {"Ne", {3, "d"}}, {2, "D", 5/2}}) // GetStateData;
Parity[state] = Even;
Wavenumber[state] = 29172.839 1/Centimeter; (* [KP08] *)
NaturalWidth[state] = 5.14 10^1 Mega/Second;  (* [KP08] *)
HyperfineA[state] = 2\[Pi] 0.1085 Mega 1/Second; (* [BHM88] *)
HyperfineB[state] = 2\[Pi] 0.179 Mega 1/Second; (* [SDJ98] *)
BranchingRatio[{"Na", 23, {"Ne", {3, "p"}}, {2, "P", 1/2}}][state] = 0;
BranchingRatio[{"Na", 23, {"Ne", {3, "p"}}, {2, "P", 3/2}}][state] = 1;

(state = {"Na", 23, {"Ne", {4, "s"}}, {2, "S", 1/2}}) // GetStateData;
Parity[state] = Even;
Wavenumber[state] = 25739.991 1/Centimeter; (* [KP08] *)
HyperfineA[state] = 2\[Pi] 202. Mega 1/Second; (* [LGH73] *)
PartialWidth[{"Na", 23, {"Ne", {3, "p"}}, {2, "P", 1/2}}][state] = 8.80 Mega/Second; (* [KP08, Hea61] *)
PartialWidth[{"Na", 23, {"Ne", {3, "p"}}, {2, "P", 3/2}}][state] = 17.6 Mega/Second; (* [KP08, Hea61] *)
NaturalWidth[state] = NaturalWidthFromPartialWidths[state];
DefineBranchingRatiosFromPartialWidths[state];

(state = {"Na", 23, {"Ne", {4, "p"}}, {2, "P", 1/2}}) // GetStateData;
Parity[state] = Odd;
Wavenumber[state] = 30266.99 1/Centimeter; (* [KP08] *)
PartialWidth[{"Na", 23, {"Ne", {3, "s"}}, {2, "S", 1/2}}][state] = (2.73 10^-2) 10^2 Mega/Second;  (* [KP08] *)
PartialWidth[{"Na", 23, {"Ne", {4, "s"}}, {2, "S", 1/2}}][state] = (6.62 10^-2) 10^2 Mega/Second;  (* [KP08] *)
PartialWidth[{"Na", 23, {"Ne", {3, "d"}}, {2, "D", 3/2}}][state] = (1.57 10^-3) 10^2 Mega/Second;  (* [KP08] *)
NaturalWidth[state] = NaturalWidthFromPartialWidths[state];
DefineBranchingRatiosFromPartialWidths[state];

(state = {"Na", 23, {"Ne", {4, "p"}}, {2, "P", 3/2}}) // GetStateData;
Parity[state] = Odd;
Wavenumber[state] = 30272.58 1/Centimeter; (* [KP08] *)
HyperfineA[state] = 2\[Pi] 6.2 Mega 1/Second; (* [SLHK70] *)
HyperfineB[state] = 2\[Pi] 1.00 Mega 1/Second; (* [SLHK70] *)
PartialWidth[{"Na", 23, {"Ne", {3, "s"}}, {2, "S", 1/2}}][state] = (2.75 10^-2) 10^2 Mega/Second;  (* [KP08] *)
PartialWidth[{"Na", 23, {"Ne", {4, "s"}}, {2, "S", 1/2}}][state] = (6.64 10^-2) 10^2 Mega/Second;  (* [KP08] *)
PartialWidth[{"Na", 23, {"Ne", {3, "d"}}, {2, "D", 3/2}}][state] = (1.59 10^-3) 10^2 Mega/Second;  (* [KP08] *)
PartialWidth[{"Na", 23, {"Ne", {3, "d"}}, {2, "D", 5/2}}][state] = (1.43 10^-3) 10^2 Mega/Second;  (* [KP08] *)
NaturalWidth[state] = NaturalWidthFromPartialWidths[state];
DefineBranchingRatiosFromPartialWidths[state];

(state = {"Na", 23, {"Ne", {4, "d"}}, {2, "D", 3/2}}) // GetStateData;
Parity[state] = Even;
Wavenumber[state] = 34548.766 1/Centimeter; (* [KP08] *)
HyperfineA[state] = 2\[Pi] 0.215 Mega 1/Second; (* [BDJM78] *)
PartialWidth[{"Na", 23, {"Ne", {3, "p"}}, {2, "P", 1/2}}][state] = 1.01 10^-1 10^2 Mega/Second; (* [KP08] *)
PartialWidth[{"Na", 23, {"Ne", {3, "p"}}, {2, "P", 3/2}}][state] = 2.02 10^-2 10^2 Mega/Second; (* [KP08] *)
PartialWidth[{"Na", 23, {"Ne", {4, "p"}}, {2, "P", 1/2}}][state] = 5.84 10^-2 10^2 Mega/Second; (* [KP08] *)
PartialWidth[{"Na", 23, {"Ne", {4, "p"}}, {2, "P", 3/2}}][state] = 1.17 10^-2 10^2 Mega/Second; (* [KP08] *)
NaturalWidth[state] = NaturalWidthFromPartialWidths[state];
DefineBranchingRatiosFromPartialWidths[state];

(state = {"Na", 23, {"Ne", {4, "d"}}, {2, "D", 5/2}}) // GetStateData;
Parity[state] = Even;
Wavenumber[state] = 34548.731 1/Centimeter; (* [KP08] *)
HyperfineA[state] = 2\[Pi] 0.029 Mega 1/Second; (* [BDJM78] *)
PartialWidth[{"Na", 23, {"Ne", {3, "p"}}, {2, "P", 3/2}}][state] = 1.21 10^-1 10^2 Mega/Second; (* [KP08] *)
PartialWidth[{"Na", 23, {"Ne", {4, "p"}}, {2, "P", 3/2}}][state] = 7.01 10^-2 10^2 Mega/Second; (* [KP08] *)
NaturalWidth[state] = NaturalWidthFromPartialWidths[state];
DefineBranchingRatiosFromPartialWidths[state];

(* K 
Natural abundance from http://physics.nist.gov/PhysRefData/Handbook/Tables/potassiumtable1.htm
Wavenumbers from http://physics.nist.gov/PhysRefData/Handbook/Tables/potassiumtable5.htm
hfs data from [AIV77]
lifetimes from [VS96]
*)

state = {"K", 39};
Abundance[state] = 0.932581;
NuclearSpin[state] = 3/2;

(state = {"K", 39, {"Ar", {4, "s"}}, {2, "S", 1/2}}) // GetStateData;
Lifetime[state] = Infinity Second;
Wavenumber[state] = 0;
HyperfineA[state] = 2\[Pi] 230.8598601 Mega 1/Second;
HyperfineB[state] = 0;
Parity[state] = Even;

(state = {"K", 39, {"Ar", {4, "p"}}, {2, "P", 1/2}}) // GetStateData;
Lifetime[state] = 26.79 Nano Second;
Wavenumber[state] = 12985.186 1/Centimeter;
HyperfineA[state] = 2\[Pi] 28.85 Mega 1/Second;
HyperfineB[state] = 0;
Parity[state] = Odd;

(state = {"K", 39, {"Ar", {4, "p"}}, {2, "P", 3/2}}) // GetStateData;
Lifetime[state] = 26.45 Nano Second;
Wavenumber[state] = 13042.896 1/Centimeter;
HyperfineA[state] = 2\[Pi] 6.06 Mega 1/Second;
HyperfineB[state] = 2\[Pi] 2.83 Mega 1/Second;
Parity[state] = Odd;

(* Rb *)

state = {"Rb", 87};
Abundance[state] = 0.27835;
NuclearSpin[state] = 3/2;

(state = {"Rb", 87, {"Kr", {5, "s"}}, {2, "S", 1/2}}) // GetStateData;
Lifetime[state] = Infinity Second;
Wavenumber[state] = 0;
HyperfineA[state] = 2\[Pi] 3417.341305452145 Mega 1/Second;
HyperfineB[state] = 0;
Parity[state] = Even;

(state = {"Rb", 87, {"Kr", {5, "p"}}, {2, "P", 1/2}}) // GetStateData;
Lifetime[state] = 27.679 Nano Second;
Wavenumber[state] = 12578.95098147 1/Centimeter;
HyperfineA[state] = 2\[Pi] 407.24 Mega 1/Second;
HyperfineB[state] = 0;
Parity[state] = Odd;

(state = {"Rb", 87, {"Kr", {5, "p"}}, {2, "P", 3/2}}) // GetStateData;
Lifetime[state] = 26.2348 Nano Second;
Wavenumber[state] = 12816.54938993 1/Centimeter;
HyperfineA[state] = 2\[Pi] 84.7185 Mega 1/Second;
HyperfineB[state] = 2\[Pi] 12.4965 Mega 1/Second;
Parity[state] = Odd;

state = {"Rb", 85};
Abundance[state] = 0.72165;
NuclearSpin[state] = 5/2;

(state = {"Rb", 85, {"Kr", {5, "s"}}, {2, "S", 1/2}}) // GetStateData;
Lifetime[state] = Infinity Second;
Wavenumber[state] = 0;
HyperfineA[state] = 2\[Pi] 1011.9108130 Mega 1/Second;
HyperfineB[state] = 0;
Parity[state] = Even;

(state = {"Rb", 85, {"Kr", {5, "p"}}, {2, "P", 1/2}}) // GetStateData;
Lifetime[state] = 27.679 Nano Second;
Wavenumber[state] = 12578.9483900 1/Centimeter;
HyperfineA[state] = 2\[Pi] 120.527 Mega 1/Second;
HyperfineB[state] = 0;
Parity[state] = Odd;

(state = {"Rb", 85, {"Kr", {5, "p"}}, {2, "P", 3/2}}) // GetStateData;
Lifetime[state] = 26.2348 Nano Second;
Wavenumber[state] = 12816.54678496 1/Centimeter;
HyperfineA[state] = 2\[Pi] 25.0020 Mega 1/Second;
HyperfineB[state] = 2\[Pi] 25.790 Mega 1/Second;
Parity[state] = Odd;

(* Cs *)

state = {"Cs", 133};
Abundance[state] = 1;
NuclearSpin[state] = 7/2;

(state = {"Cs", 133, {"Xe", {6, "s"}}, {2, "S", 1/2}}) // GetStateData;
Lifetime[state] = Infinity Second;
Wavenumber[state] = 0;
HyperfineA[state] = 2\[Pi] 2298.1579425 Mega 1/Second;
HyperfineB[state] = 0;
Parity[state] = Even;

(state = {"Cs", 133, {"Xe", {6, "p"}}, {2, "P", 1/2}}) // GetStateData;
Lifetime[state] = 34.791 Nano Second;
Wavenumber[state] = 11178.2681607 1/Centimeter;
HyperfineA[state] = 2\[Pi] 291.9201 Mega 1/Second;
HyperfineB[state] = 0;
Parity[state] = Odd;

(state = {"Cs", 133, {"Xe", {6, "p"}}, {2, "P", 3/2}}) // GetStateData;
Lifetime[state] = 30.405 Nano Second;
Wavenumber[state] = 11732.3071049 1/Centimeter;
HyperfineA[state] = 2\[Pi] 50.28827 Mega 1/Second;
HyperfineB[state] = 2\[Pi] (-0.4934) Mega 1/Second;
HyperfineC[state] = 2\[Pi] 0.000560 Mega 1/Second;
Parity[state] = Odd;

(* Lu 177 *)

state = {"Lu", 177};
Abundance[state] = 0;
NuclearSpin[state] = 7/2;

(state = {"Lu", 177, {"Xe", {4, "f", 14}, {5, "d"}, {6, "s", 2}}, {2, "D", 3/2}}) // GetStateData;
Lifetime[state] = Infinity Second;
Wavenumber[state] = 0;
HyperfineA[state] = 194.84 Mega 2 Pi/Second; (*[PS62]*)
HyperfineB[state] = 1466.71 Mega 2 Pi/Second; (*[PS62]*)
Parity[state] = Even;

(state = {"Lu", 177, {"Xe", {4, "f", 14}, {5, "d"}, {6, "s"}, {6, "p"}}, {2, "D", 3/2}}) // GetStateData;
Lifetime[state] = 43.9 Nano Second; (*[FDHL+00]*)
Wavenumber[state] = 22124.76 1/Centimeter;
HyperfineA[state] = -1428.9 Mega 2 Pi/Second; (*[GBK+98]*)
HyperfineB[state] = -484. Mega 2 Pi/Second; (*[GBK+98]*)
Parity[state] = Odd;

(* Lu 177m *)

state = {"Lu", {177, "m"}};
Abundance[state] = 0;
NuclearSpin[state] = 23/2;

(state = {"Lu", {177, "m"}, {"Xe", {4, "f", 14}, {5, "d"}, {6, "s", 2}}, {2, "D", 3/2}}) // GetStateData;
Lifetime[state] = Infinity Second;
Wavenumber[state] = 0;
HyperfineA[state] = 61.2 Mega 2 Pi/Second; (*[GBK+98]*)
HyperfineB[state] = 2472. Mega 2 Pi/Second; (*[GBK+98]*)
Parity[state] = Even;

(state = {"Lu", {177, "m"}, {"Xe", {4, "f", 14}, {5, "d"}, {6, "s"}, {6, "p"}}, {2, "D", 3/2}}) // GetStateData;
Lifetime[state] = 43.9 Nano Second; (*[FDHL+00]*)
Wavenumber[state] = 22124.76 1/Centimeter;
HyperfineA[state] = -448.0 Mega 2 Pi/Second; (*[GBK+98]*)
HyperfineB[state] = -807. Mega 2 Pi/Second; (*[GBK+98]*)
Parity[state] = Odd;

(* Cu 63 *)
(* HFS data from:
	Spectrochimica Acta Part B: Atomic Spectroscopy Volume 48, Issue 10, August 1993, Pages 1259-1264  
	Phys. Rev. 105, 581-588 (1957), 
	Optics Communications Volume 31, Issue 1, October 1979, Pages 28-30   
   Wavenumber data from NIST website
   Lifetime data from L J Curtis et al 1976 Phys. Scr. 13 109,
   Abundance data from http://www.sisweb.com/referenc/source/exactmas.htm
   *)

state = {"Cu", 63};
Abundance[state] = 0.6917;
NuclearSpin[state] = 3/2;

(state = {"Cu", 63, {"Ar", {3, "d", 10}, {4, "s"}}, {2, "S", 1/2}}) // GetStateData;
Lifetime[state] = Infinity Second;
Wavenumber[state] = 0;
HyperfineA[state] = 2\[Pi] 5866.92 Mega 1/Second;
HyperfineB[state] = 0;

(state = {"Cu", 63, {"Ar", {3, "d", 10}, {4, "p"}}, {2, "P", 1/2}}) // GetStateData;
Lifetime[state] = 7.6 Nano Second;
Wavenumber[state] = 30535.3 1/Centimeter;
HyperfineA[state] = 2\[Pi] 505.6 Mega 1/Second;
HyperfineB[state] = 0;

(state = {"Cu", 63, {"Ar", {3, "d", 10}, {4, "p"}}, {2, "P", 3/2}}) // GetStateData;
Lifetime[state] = 7.6 Nano Second;
Wavenumber[state] = 30783.7 1/Centimeter;
HyperfineA[state] = 2\[Pi] 194.9 Mega 1/Second;
HyperfineB[state] = 2\[Pi] (-29.1) Mega 1/Second;

(state = {"Cu", 63, {"Ar", {3, "d", 9}, {4, "s", 2}}, {2, "D", 3/2}}) // GetStateData;
Wavenumber[state] = 13245.4 1/Centimeter;
HyperfineA[state] = 2\[Pi] Convert[SpeedOfLight 0.0614 1/Centimeter, Mega 1/Second];
HyperfineB[state] = 2\[Pi] Convert[SpeedOfLight 1.88 10^(-4) 1/Centimeter, Mega 1/Second];

(state = {"Cu", 63, {"Ar", {3, "d", 9}, {4, "s", 2}}, {2, "D", 5/2}}) // GetStateData;
Wavenumber[state] = 11202.6 1/Centimeter;
HyperfineA[state] = 2\[Pi] Convert[SpeedOfLight 0.0249 1/Centimeter, Mega 1/Second];
HyperfineB[state] = 2\[Pi] Convert[SpeedOfLight 9 10^(-5) 1/Centimeter, Mega 1/Second];

(* Cu 65*)

state = {"Cu", 65};
Abundance[state] = 0.3083;
NuclearSpin[state] = 3/2;

(state = {"Cu", 65, {"Ar", {3, "d", 10}, {4, "s"}}, {2, "S", 1/2}}) // GetStateData;
Lifetime[state] = Infinity Second;
Wavenumber[state] = 0;
HyperfineA[state] = 2\[Pi] 6284.41 Mega 1/Second;
HyperfineB[state] = 0;

(state = {"Cu", 65, {"Ar", {3, "d", 10}, {4, "p"}}, {2, "P", 1/2}}) // GetStateData;
Lifetime[state] = 7.6 Nano Second;
Wavenumber[state] = 30535.3 1/Centimeter;
HyperfineA[state] = 2\[Pi] 542.9 Mega 1/Second;
HyperfineB[state] = 0;

(state = {"Cu", 65, {"Ar", {3, "d", 10}, {4, "p"}}, {2, "P", 3/2}}) // GetStateData;
Lifetime[state] = 7.6 Nano Second;
Wavenumber[state] = 30783.7 1/Centimeter;
HyperfineA[state] = 2\[Pi] 207.9 Mega 1/Second;
HyperfineB[state] = 2\[Pi] (-26.9) Mega 1/Second;

(state = {"Cu", 65, {"Ar", {3, "d", 9}, {4, "s", 2}}, {2, "D", 3/2}}) // GetStateData;
Wavenumber[state] = 13245.4 1/Centimeter;
HyperfineA[state] = 2\[Pi] Convert[SpeedOfLight 0.0656 1/Centimeter, Mega 1/Second];
HyperfineB[state] = 2\[Pi] Convert[SpeedOfLight 9 10^(-5) 1/Centimeter, Mega 1/Second];

(state = {"Cu", 65, {"Ar", {3, "d", 9}, {4, "s", 2}}, {2, "D", 5/2}}) // GetStateData;
Wavenumber[state] = 11202.6 1/Centimeter;
HyperfineA[state] = 2\[Pi] Convert[SpeedOfLight 0.0252 1/Centimeter, Mega 1/Second];
HyperfineB[state] = 2\[Pi] Convert[SpeedOfLight 3.7 10^(-5) 1/Centimeter, Mega 1/Second];

AtomicTransition[atom_String, line_String] := Switch[{atom, line},
    {"H", "D1"},
    {{"H", 1, {{1, "s"}}, {2, "S", 1/2}}, {"H", 1, {{2, "p"}}, {2, "P", 1/2}}},
    {"H", "D2"},
    {{"H", 1, {{1, "s"}}, {2, "S", 1/2}}, {"H", 1, {{2, "p"}}, {2, "P", 3/2}}},
    {"Na"|"Na23", "D1"},
    {{"Na", 23, {"Ne", {3, "s"}}, {2, "S", 1/2}}, {"Na", 23, {"Ne", {3, "p"}}, {2, "P", 1/2}}},
    {"Na"|"Na23", "D2"},
    {{"Na", 23, {"Ne", {3, "s"}}, {2, "S", 1/2}}, {"Na", 23, {"Ne", {3, "p"}}, {2, "P", 3/2}}},
    {"K"|"K39", "D1"},
    {{"K", 39, {"Ar", {4, "s"}}, {2, "S", 1/2}}, {"K", 39, {"Ar", {4, "p"}}, {2, "P", 1/2}}},
    {"K"|"K39", "D2"},
    {{"K", 39, {"Ar", {4, "s"}}, {2, "S", 1/2}}, {"K", 39, {"Ar", {4, "p"}}, {2, "P", 3/2}}},
    {"Rb87", "D1"},
    {{"Rb", 87, {"Kr", {5, "s"}}, {2, "S", 1/2}}, {"Rb", 87, {"Kr", {5, "p"}}, {2, "P", 1/2}}},
    {"Rb87", "D2"},
    {{"Rb", 87, {"Kr", {5, "s"}}, {2, "S", 1/2}}, {"Rb", 87, {"Kr", {5, "p"}}, {2, "P", 3/2}}},
    {"Rb85", "D1"},
    {{"Rb", 85, {"Kr", {5, "s"}}, {2, "S", 1/2}}, {"Rb", 85, {"Kr", {5, "p"}}, {2, "P", 1/2}}},
    {"Rb85", "D2"},
    {{"Rb", 85, {"Kr", {5, "s"}}, {2, "S", 1/2}}, {"Rb", 85, {"Kr", {5, "p"}}, {2, "P", 3/2}}},
    {"Cs"|"Cs133", "D1"},
    {{"Cs", 133, {"Xe", {6, "s"}}, {2, "S", 1/2}}, {"Cs", 133, {"Xe", {6, "p"}}, {2, "P", 1/2}}},
    {"Cs"|"Cs133", "D2"},
    {{"Cs", 133, {"Xe", {6, "s"}}, {2, "S", 1/2}}, {"Cs", 133, {"Xe", {6, "p"}}, {2, "P", 3/2}}},
    _,
    Print["Invalid atom or transition."];Abort[];
];

If[AtomicDataDefinedQ[#, Wavenumber], Energy@# = EnergyFromWavenumber@#]& /@ AtomicData[];

If[AtomicDataDefinedQ[#, Lifetime] && !AtomicDataDefinedQ[#, NaturalWidth],
    NaturalWidth@# = NaturalWidthFromLifetime@#
]& /@ AtomicData[];

If[AtomicDataDefinedQ[#, NaturalWidth] && !AtomicDataDefinedQ[#, Lifetime],
    Lifetime@# = LifetimeFromNaturalWidth@#
]& /@ AtomicData[];

End[];

EndPackage[];
